@@ -7,7 +7,7 @@
 #
 # $Id$
 #
-# Copyright B. Cornec 2007-2012
+# Copyright B. Cornec 2007-2016
 # Eric Anderson's changes are (c) Copyright 2012 Hewlett Packard
 # Provided under the GPL v2
 
@@ -15,6 +15,7 @@ package ProjectBuilder::VCS;
 
 use strict 'vars';
 use Carp 'confess';
+use Cwd 'abs_path';
 use Data::Dumper;
 use English;
 use File::Basename;
@@ -34,7 +35,7 @@ use Exporter;
 # any code which uses this module.
  
 our @ISA = qw(Exporter);
-our @EXPORT = qw(pb_vcs_export pb_vcs_get_uri pb_vcs_copy pb_vcs_checkout pb_vcs_up pb_vcs_checkin pb_vcs_isdiff pb_vcs_add pb_vcs_cmd);
+our @EXPORT = qw(pb_vcs_export pb_vcs_get_uri pb_vcs_copy pb_vcs_checkout pb_vcs_up pb_vcs_checkin pb_vcs_isdiff pb_vcs_add pb_vcs_add_if_not_in pb_vcs_cmd pb_vcs_compliant);
 ($VERSION,$REVISION) = pb_version_init();
 
 =pod
@@ -111,11 +112,11 @@ if ($scheme =~ /^svn/) {
 		}
 	} else {
 		# Look at svk admin hotcopy
-		die "Unable to export from svk without a source defined";
+		confess "Unable to export from svk without a source defined";
 	}
 } elsif ($scheme eq "dir") {
 	pb_system("cp -r $path $destdir","Copying $uri from DIR to $destdir ");
-} elsif (($scheme eq "http") || ($scheme eq "ftp")) {
+} elsif (($scheme =~ /http/) || ($scheme eq "ftp")) {
 	my $f = basename($path);
 	unlink "$ENV{'PBTMP'}/$f";
 	pb_system("$vcscmd $ENV{'PBTMP'}/$f $uri","Downloading $uri with $vcscmd to $ENV{'PBTMP'}/$f\n");
@@ -130,7 +131,7 @@ if ($scheme =~ /^svn/) {
 	};
 	if ($@) {
 		# File::MimeInfo not found
-		die("ERROR: Install File::MimeInfo to handle scheme $scheme\n");
+		confess("ERROR: Install File::MimeInfo to handle scheme $scheme\n");
 	}
 
 	my $mm = mimetype($path);
@@ -183,17 +184,39 @@ if ($scheme =~ /^svn/) {
 		pb_system("$vcscmd clone $uri $destdir","Exporting $uri from Mercurial to $destdir ");
 	}
 } elsif ($scheme =~ /^git/) {
-	if (defined $source) {
-		if (-d $source) {
-			$tmp = $destdir;
+	if ($scheme =~ /svn/) {
+		if (defined $source) {
+			if (-d $source) {
+				$tmp = $destdir;
+			} else {
+				$tmp = "$destdir/".basename($source);
+			}
+			$source = pb_vcs_mod_htftp($source,"git");
+			pb_system("cp -a $source $tmp","Exporting $source from $scheme to $tmp ");
 		} else {
-			$tmp = "$destdir/".basename($source);
+			$uri = pb_vcs_mod_htftp($uri,"git");
+			pb_system("$vcscmd clone $uri $destdir","Exporting $uri from $scheme to $destdir ");
 		}
-		$source = pb_vcs_mod_htftp($source,"git");
-		pb_system("cd $source ; $vcscmd archive --format=tar HEAD | (mkdir $tmp && cd $tmp && tar xf -)","Exporting $source/HEAD from GIT to $tmp ");
 	} else {
-		$uri = pb_vcs_mod_htftp($uri,"git");
-		pb_system("$vcscmd clone $uri $destdir","Exporting $uri from GIT to $destdir ");
+		if (defined $source) {
+			if (-d $source) {
+				$tmp = $destdir;
+			} else {
+				$tmp = "$destdir/".basename($source);
+			}
+			$source = pb_vcs_mod_htftp($source,"git");
+			my ($pbpbr) = pb_conf_get_if("pbpbr");
+			if ((defined $pbpbr) && (defined $pbpbr->{$ENV{'PBPROJ'}})) {
+				# The project uses pbr so benefit from it to export data
+				pb_system("cd $source ; mkdir $tmp ; python setup.py sdist --keep-temp --dist-dir $tmp ; cd $tmp ; file=`ls *.tar.gz` ;  if [ _\$file = _ ] || [ ! -f \$file ]; then exit -1; fi; tar xfz \$file ; dir=`tar tvfz \$file | head -1 | awk '{print \$6}'` ; if [ ! -d \$dir ] || [ _\$dir = _ ] || [ \$dir = / ]; then exit -1 ; fi ; mv \$dir/* \$dir/.??* . ; rmdir \$dir ; rm -f \$file ; ls -al ","Exporting current $source from GIT with pbr to $tmp ");
+			} else {
+				# no pbr do it ourselves
+				pb_system("cd $source ; stid=`$vcscmd stash create` ; $vcscmd archive --format=tar \$\{stid:=HEAD\} | (mkdir $tmp && cd $tmp && tar xf -)","Exporting current $source from GIT to $tmp ");
+			}
+		} else {
+			$uri = pb_vcs_mod_htftp($uri,"git");
+			pb_system("$vcscmd clone $uri $destdir","Exporting $uri from GIT to $destdir ");
+		}
 	}
 } elsif ($scheme =~ /^cvs/) {
 	# CVS needs a relative path !
@@ -219,7 +242,7 @@ if ($scheme =~ /^svn/) {
 	}
 	pb_system("cd $dir ; $vcscmd -d $account\@$host:$path export $cvsopt -d $base $tmp1","Exporting $tmp1 from $source under CVS to $destdir ");
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 return(undef);
 }
@@ -242,7 +265,7 @@ my $void = "";
 my $vcscmd = pb_vcs_cmd($scheme);
 
 if ($scheme =~ /^svn/) {
-	open(PIPE,"LANGUAGE=C $vcscmd info $dir |") || return("");
+	open(PIPE,"LANGUAGE=C $vcscmd info $dir 2> /dev/null |") || return("");
 	while (<PIPE>) {
 		($void,$res) = split(/^URL:/) if (/^URL:/);
 	}
@@ -250,7 +273,7 @@ if ($scheme =~ /^svn/) {
 	close(PIPE);
 	chomp($res);
 } elsif ($scheme =~ /^svk/) {
-	open(PIPE,"LANGUAGE=C $vcscmd info $dir |") || return("");
+	open(PIPE,"LANGUAGE=C $vcscmd info $dir 2> /dev/null |") || return("");
 	my $void2 = "";
 	while (<PIPE>) {
 		($void,$void2,$res) = split(/ /) if (/^Depot/);
@@ -266,17 +289,42 @@ if ($scheme =~ /^svn/) {
 	close(HGRC);
 	chomp($res);
 } elsif ($scheme =~ /^git/) {
-	open(GIT,"git --git-dir=$dir/.git remote -v |") || return("");
-	while (<GIT>) {
-		next unless (/^origin\s+(\S+) \(push\)$/);
-		return $1;
+	if ($scheme =~ /svn/) {
+		my $cwd = abs_path();
+		chdir($dir) || return("");;
+		open(PIPE,"LANGUAGE=C $vcscmd info . 2> /dev/null |") || return("");
+		chdir($cwd) || return("");
+		while (<PIPE>) {
+			($void,$res) = split(/^URL:/) if (/^URL:/);
+		}
+		$res =~ s/^\s*//;
+		close(PIPE);
+		chomp($res);
+		# We've got an SVN ref so add git in front of it for coherency
+		$res = "git+".$res;
+	} else {
+		# Pure git
+		# First we may deal with a separate git repo under $dir
+		if ( -d "$dir/.git" ) {
+			open(GIT,"LANGUAGE=C $vcscmd --git-dir=$dir/.git remote -v 2> /dev/null |") || return("");
+		} else {
+			# If not, the pbconf dir may be in the pbprojdir so sharing the .git dir
+			my $cwd = abs_path();
+			chdir($dir) || return("");;
+			open(GIT,"LANGUAGE=C $vcscmd remote -v 2> /dev/null |") || return("");
+			chdir($cwd) || return("");
+		}
+		while (<GIT>) {
+			next unless (/^origin\s+(\S+) \(push\)$/);
+			return $1;
+		}
+		close(GIT);
+		warn "Unable to find remote origin for $dir";
+		return "";
 	}
-	close(GIT);
-	warn "Unable to find origin remote for $dir";
-	return "";
 } elsif ($scheme =~ /^cvs/) {
 	# This path is always the root path of CVS, but we may be below
-	open(FILE,"$dir/CVS/Root") || die "$dir isn't CVS controlled";
+	open(FILE,"$dir/CVS/Root") || confess "$dir isn't CVS controlled";
 	$res = <FILE>;
 	chomp($res);
 	close(FILE);
@@ -285,7 +333,7 @@ if ($scheme =~ /^svn/) {
 	while ((! -d "$rdir/CVSROOT") && ($rdir ne "/")) {
 		$rdir = dirname($rdir);
 	}
-	die "Unable to find a CVSROOT dir in the parents of $dir" if (! -d "$rdir/CVSROOT");
+	confess "Unable to find a CVSROOT dir in the parents of $dir" if (! -d "$rdir/CVSROOT");
 	#compute our place under that root dir - should be a relative path
 	$dir =~ s|^$rdir||;
 	my $suffix = "";
@@ -299,7 +347,7 @@ if ($scheme =~ /^svn/) {
 	}
 	$res = $prefix.$res.$suffix;
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 pb_log(1,"pb_vcs_get_uri returns $res\n");
 return($res);
@@ -328,10 +376,10 @@ if ($scheme =~ /^svn/) {
 	$oldurl = pb_vcs_mod_htftp($oldurl,"svn");
 	$newurl = pb_vcs_mod_htftp($newurl,"svn");
 	pb_system("$vcscmd copy -m \"Creation of $newurl from $oldurl\" $oldurl $newurl","Copying $oldurl to $newurl ");
-} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(file)\b/o) {
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
 	# Nothing to do.
 } else {
-	die "cms $scheme unknown for project management";
+	confess "cms $scheme unknown for project management";
 }
 }
 
@@ -363,7 +411,7 @@ if ($scheme =~ /^svn/) {
 } elsif ($scheme =~ /^git/) {
 	$url = pb_vcs_mod_htftp($url,"git");
 	pb_system("$vcscmd clone $url $destination","Checking out $url to $destination ");
-} elsif (($scheme eq "ftp") || ($scheme eq "http")) {
+} elsif (($scheme eq "ftp") || ($scheme =~ /http/)) {
 	return;
 } elsif ($scheme =~ /^cvs/) {
 	my ($scheme, $account, $host, $port, $path) = pb_get_uri($url);
@@ -384,7 +432,7 @@ if ($scheme =~ /^svn/) {
 } elsif ($scheme =~ /^file/) {
 	pb_vcs_export($url,undef,$destination);
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 }
 
@@ -405,12 +453,12 @@ if ($scheme =~ /^((svn)|(cvs)|(svk))/o) {
 	pb_system("$vcscmd up ".join(' ',@dir),"Updating ".join(' ',@dir));
 } elsif ($scheme =~ /^((hg)|(git))/o) {
 	foreach my $d (@dir) {
-		pb_system("(cd $d && $vcscmd pull)", "Updating $d ");
+		pb_system("(cd $d && $vcscmd fetch)", "Updating $d ");
 	}
-} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(file)\b/o) {
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
 	# Nothing to do.
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 }
 
@@ -433,14 +481,39 @@ if ($scheme =~ /^((svn)|(cvs)|(svk))/o) {
 	pb_system("cd $dir && $vcscmd ci -m \"$msg\" .","Checking in $dir ");
 } elsif ($scheme =~ /^git/) {
 	pb_system("cd $dir && $vcscmd commit -a -m \"$msg\"", "Checking in $dir ");
-} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(file)\b/o) {
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
 	# Nothing to do.
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 pb_vcs_up($scheme,$dir);
 }
 
+=item B<pb_vcs_add_if_not_in>
+
+This function adds to a VCS content from a local directory if the content wasn't already managed under th VCS.
+The first parameter is the schema of the VCS systems (svn, cvs, svn+ssh, ...)
+The second parameter is a list of directory/file to add.
+
+=cut
+
+sub pb_vcs_add_if_not_in {
+my $scheme = shift;
+my @f = @_;
+my $vcscmd = pb_vcs_cmd($scheme);
+
+if ($scheme =~ /^((hg)|(git)|(svn)|(svk)|(cvs))/o) {
+	for my $f (@f) {
+		my $uri = pb_vcs_get_uri($scheme,$f);
+		pb_vcs_add($scheme,$f) if ($uri !~ /^$scheme/);
+	}
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
+	# Nothing to do.
+} else {
+	confess "cms $scheme unknown";
+}
+}
+
 =item B<pb_vcs_add>
 
 This function adds to a VCS content from a local directory.
@@ -456,17 +529,17 @@ my $vcscmd = pb_vcs_cmd($scheme);
 
 if ($scheme =~ /^((hg)|(git)|(svn)|(svk)|(cvs))/o) {
 	pb_system("$vcscmd add ".join(' ',@f),"Adding ".join(' ',@f)." to VCS ");
-} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(file)\b/o) {
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
 	# Nothing to do.
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 pb_vcs_up($scheme,@f);
 }
 
 =item B<pb_vcs_isdiff>
 
-This function returns a integer indicating the number f differences between the VCS content and the local directory where it's checked out.
+This function returns a integer indicating the number of differences between the VCS content and the local directory where it's checked out.
 The first parameter is the schema of the VCS systems (svn, cvs, svn+ssh, ...)
 The second parameter is the directory to consider.
 
@@ -478,18 +551,18 @@ my $dir =shift;
 my $vcscmd = pb_vcs_cmd($scheme);
 my $l = undef;
 
-if ($scheme =~ /^((svn)|(cvs)|(svk))/o) {
-	open(PIPE,"$vcscmd diff $dir |") || die "Unable to get $vcscmd diff from $dir";
+if ($scheme =~ /^((svn)|(cvs)|(svk)|(git))/o) {
+	open(PIPE,"$vcscmd diff $dir |") || confess "Unable to get $vcscmd diff from $dir";
 	$l = 0;
 	while (<PIPE>) {
 		# Skipping normal messages in case of CVS
 		next if (/^cvs diff:/);
 		$l++;
 	}
-} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(file)\b/o) {
+} elsif ($scheme =~ /^(flat)|(ftp)|(http)|(https)|(file)\b/o) {
 	$l = 0;
 } else {
-	die "cms $scheme unknown";
+	confess "cms $scheme unknown";
 }
 pb_log(1,"pb_vcs_isdiff returns $l\n");
 return($l);
@@ -533,33 +606,135 @@ if (defined $ENV{'PBVCSOPT'}) {
 }
 
 if ($scheme =~ /hg/) {
-	return($cmd."hg".$cmdopt)
+	$cmd .= "hg".$cmdopt;
 } elsif ($scheme =~ /git/) {
-	return($cmd."git".$cmdopt)
+	if ($scheme =~ /svn/) {
+		$cmd .= "git svn".$cmdopt;
+	} else {
+		$cmd .= "git".$cmdopt;
+	}
 } elsif ($scheme =~ /svn/) {
-	return($cmd."svn".$cmdopt)
+	$cmd .= "svn".$cmdopt;
 } elsif ($scheme =~ /svk/) {
-	return($cmd."svk".$cmdopt)
+	$cmd .= "svk".$cmdopt;
 } elsif ($scheme =~ /cvs/) {
-	return($cmd."cvs".$cmdopt)
+	$cmd .= "cvs".$cmdopt;
 } elsif (($scheme =~ /http/) || ($scheme =~ /ftp/)) {
 	my $command = pb_check_req("wget",1);
 	if (-x $command) {
-		return($cmd."$command -nv -O ");
+		$cmd .= "$command -nv -O ";
 	} else {
 		$command = pb_check_req("curl",1);
 		if (-x $command) {
-			return($cmd."$command -o ");
+			$cmd .= "$command -o ";
 		} else {
-			die "Unable to handle $scheme.\nNo wget/curl available, please install one of those";
+			confess "Unable to handle $scheme.\nNo wget/curl available, please install one of those";
 		}
 	}
 } else {
-	return($cmd);
+	$cmd = "";
 }
+pb_log(3,"pb_vcs_cmd returns $cmd\n");
+return($cmd);
+}
+
+=item B<pb_vcs_compliant>
+
+This function checks the compliance of the project and the pbconf directory.
+The first parameter is the key name of the value that needs to be read in the configuration file.
+The second parameter is the environment variable this key will populate.
+The third parameter is the location of the pbconf dir.
+The fourth parameter is the URI of the CMS content related to the pbconf dir.
+The fifth parameter indicates whether we should inititate the context or not.
+
+=cut
+
+sub pb_vcs_compliant {
+
+my $param = shift;
+my $envar = shift;
+my $defdir = shift;
+my $uri = shift;
+my $pbinit = shift;
+my %pdir;
+
+pb_log(1,"pb_vcs_compliant: envar: $envar - defdir: $defdir - uri: $uri\n");
+my ($pdir) = pb_conf_get_if($param) if (defined $param);
+if (defined $pdir) {
+	%pdir = %$pdir;
+}
+
+if ((defined $pdir) && (%pdir) && (defined $pdir{$ENV{'PBPROJ'}})) {
+	# That's always the environment variable that will be used
+	$ENV{$envar} = $pdir{$ENV{'PBPROJ'}};
+} else {
+	if (defined $param) {
+		pb_log(1,"WARNING: no $param defined, using $defdir\n");
+		pb_log(1,"         Please create a $param reference for project $ENV{'PBPROJ'} in $ENV{'PBETC'}\n");
+		pb_log(1,"         if you want to use another directory\n");
+	}
+	$ENV{$envar} = "$defdir";
+}
+
+# Expand potential env variable in it
+eval { $ENV{$envar} =~ s/(\$ENV.+\})/$1/eeg };
+pb_log(2,"$envar: $ENV{$envar}\n");
+
+my ($scheme, $account, $host, $port, $path) = pb_get_uri($uri);
+
+if (($scheme !~ /^cvs/) && ($scheme !~ /^svn/) && ($scheme !~ /^svk/) && ($scheme !~ /^hg/) && ($scheme !~ /^git/)) {
+	# Do not compare if it's not a real cms
+	pb_log(1,"pb_vcs_compliant useless\n");
+	return;
+} elsif (defined $pbinit) {
+	pb_mkdir_p("$ENV{$envar}");
+} elsif (! -d "$ENV{$envar}") {
+	# Either we have a version in the uri, and it should be the same
+	# as the one in the envar. Or we should add the version to the uri
+	# But not if it's git as it manages version branches internally
+	if ((basename($uri) ne basename($ENV{$envar})) && ($scheme !~ /^git/)) {
+		$uri .= "/".basename($ENV{$envar})
+	}
+	pb_log(1,"Checking out $uri\n");
+	# Create structure and remove end dir before exporting
+	pb_mkdir_p("$ENV{$envar}");
+	pb_rm_rf($ENV{$envar});
+	pb_vcs_checkout($scheme,$uri,$ENV{$envar});
+} else {
+	pb_log(1,"$uri found locally, checking content\n");
+	my $cmsurl = pb_vcs_get_uri($scheme,$ENV{$envar});
+	my ($scheme2, $account2, $host2, $port2, $path2) = pb_get_uri($cmsurl);
+	# For svk, scheme doesn't appear in svk info so remove it here in uri coming from conf file 
+	# which needs it to trigger correct behaviour
+	$uri =~ s/^svk://;
+	if ($scheme2 =~ /^git/) {
+		# remove schema from `git+file:` and `git+dir:` urls
+		# TODO: handle query-parameters
+		$uri =~ s/^git\+(file|dir|ssh):[\/]*//;
+		# Expand potential env variable in it -- this is required due to the consistency check
+		$uri =~ s/(\$ENV.+\})/$1/eeg;
+	} elsif ($scheme2 =~ /^hg/) {
+		# This VCS manages branches internally not with different tree structures
+		# Assuming it's correct for now.
+		return;
+	}
+	# Remove git+ part if only in scheme
+	$uri =~ s/^git\+// if (($scheme =~ /^git\+/) && ($scheme2 !~ /^git\+/));
+
+	if ($cmsurl ne $uri) {
+		# The local content doesn't correpond to the repository
+		pb_log(0,"ERROR: Inconsistency detected:\n");
+		pb_log(0,"       * $ENV{$envar} ($envar) refers to $cmsurl but\n");
+		pb_log(0,"       * $ENV{'PBETC'} refers to $uri\n");
+		die "Project $ENV{'PBPROJ'} is not Project-Builder compliant.";
+	} else {
+		pb_log(1,"Content correct - doing nothing - you may want to update your repository however\n");
+		# they match - do nothing - there may be local changes
+	}
+}
+pb_log(1,"pb_vcs_compliant end\n");
 }
 
-	
 
 =back 
 
